from __future__ import absolute_import, division, print_function, with_statement

CLOSED = 'CLOSED'
AWAITING_AUTHENTICATION = 'AWAITING_AUTHENTICATION'
AWAITING_CONNECTION = 'AWAITING_CONNECTION'
AUTHENTICATING = 'AUTHENTICATING'
OPEN = 'OPEN'
ERROR = 'ERROR'
RECONNECTING = 'RECONNECTING'
CHALLENGING = 'CHALLENGING'



CONNECTION_ERROR = 'connectionError'
CONNECTION_STATE_CHANGED = 'connectionStateChanged'
MAX_RECONNECTION_ATTEMPTS_REACHED = 'MAX_RECONNECTION_ATTEMPTS_REACHED'
CONNECTION_AUTHENTICATION_TIMEOUT = 'CONNECTION_AUTHENTICATION_TIMEOUT'
ACK_TIMEOUT = 'ACK_TIMEOUT'
NO_RPC_PROVIDER = 'NO_RPC_PROVIDER'
RESPONSE_TIMEOUT = 'RESPONSE_TIMEOUT'
DELETE_TIMEOUT = 'DELETE_TIMEOUT'
UNSOLICITED_MESSAGE = 'UNSOLICITED_MESSAGE'
MESSAGE_DENIED = 'MESSAGE_DENIED'
MESSAGE_PARSE_ERROR = 'MESSAGE_PARSE_ERROR'
VERSION_EXISTS = 'VERSION_EXISTS'
NOT_AUTHENTICATED = 'NOT_AUTHENTICATED'
MESSAGE_PERMISSION_ERROR = 'MESSAGE_PERMISSION_ERROR'
LISTENER_EXISTS = 'LISTENER_EXISTS'
NOT_LISTENING = 'NOT_LISTENING'
TOO_MANY_AUTH_ATTEMPTS = 'TOO_MANY_AUTH_ATTEMPTS'
IS_CLOSED = 'IS_CLOSED'
RECORD_NOT_FOUND = 'RECORD_NOT_FOUND'
NOT_SUBSCRIBED = 'NOT_SUBSCRIBED'
INVALID_RPC_CORRELATION_ID = 'INVALID_RPC_CORRELATION_ID'



"""Deepstream event handling."""

from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.constants import actions
from deepstreampy.constants import topic as topic_constants
from deepstreampy.constants import event as event_constants
from deepstreampy.message import message_parser
from deepstreampy.message import message_builder
from deepstreampy.utils import Listener
from deepstreampy.utils import AckTimeoutRegistry
from deepstreampy.utils import ResubscribeNotifier

from tornado import concurrent

from pyee import EventEmitter


class EventHandler(object):
    """Handles incoming and outgoing messages related to deepstream events.
    """

    def __init__(self, connection, client, **options):
        self._options = options
        self._connection = connection
        self._client = client
        self._emitter = EventEmitter()
        self._listener = {}

        subscription_timeout = options.get("subscriptionTimeout", 15)
        self._ack_timeout_registry = AckTimeoutRegistry(client,
                                                        topic_constants.EVENT,
                                                        subscription_timeout)
        self._resubscribe_notifier = ResubscribeNotifier(client,
                                                         self._resubscribe)

    def subscribe(self, name, callback):
        """Subscribe to an event.

        Adds a callback for both locally emited events as well as events emitted
        by other clients.

        Args:
            name (str): The name of the event.
            callback (callable): The function to call when an event is received.

        """
        future = None
        if not self._emitter.listeners(name):
            self._ack_timeout_registry.add(name, actions.SUBSCRIBE)
            future = self._connection.send_message(topic_constants.EVENT,
                                                   actions.SUBSCRIBE,
                                                   [name])
        else:
            future = concurrent.Future()
            future.set_result(None)

        self._emitter.on(name, callback)

        return future

    def unsubscribe(self, name, callback):
        """Unsubscribe from an event.

        Removes the callback for the specified event, and notifies the server
        of the change.

        Args:
            name (str): The name of the event
            callback (callable): The callback to remove

        """
        self._emitter.remove_listener(name, callback)

        if not self._emitter.listeners(name):
            self._ack_timeout_registry.add(name, actions.UNSUBSCRIBE)
            return self._connection.send_message(topic_constants.EVENT,
                                                 actions.UNSUBSCRIBE,
                                                 [name])

        future = concurrent.Future()
        future.set_result(None)
        return future

    def emit(self, name, data):
        """Emit an event locally, and tell the server to broadcast it.

        Other connected clients will also receive the event.

        Args:
            name (str): The name of the event.
            data: JSON serializable data to send along with the event.

        """
        future = self._connection.send_message(
            topic_constants.EVENT, actions.EVENT,
            [name, message_builder.typed(data)])

        self._emitter.emit(name, data)

        return future

    def listen(self, pattern, callback):
        """Register as listener for event subscriptions from other clients.

        Args:
            pattern (str): Regular expression pattern to match subscriptions to
            callback (callable): A function that will be called when an event
                has been initially subscribed to or is no longer subscribed.

                Expects the following arguments:
                    event_name (str)
                    is_subscribed (bool)
                    response (callable, callable)

        """
        if (pattern in self._listener and
                not self._listener[pattern].destroy_pending):
            self._client._on_error(topic_constants.EVENT,
                                   event_constants.LISTENER_EXISTS,
                                   pattern)
            future = concurrent.Future()
            future.set_result(None)
            return future

        elif pattern in self._listener:
            self._listener[pattern].destroy()

        listener = Listener(topic_constants.EVENT,
                            pattern,
                            callback,
                            self._options,
                            self._client,
                            self._connection)
        self._listener[pattern] = listener

        return listener.send_future

    def unlisten(self, pattern):
        """Stop listening to the specified pattern.

        Remove a previously registered listening pattern. The client will no
        longer be listening for active/inactive subscriptions.

        Args:
            pattern: The regular expression pattern to remove
        """
        if pattern not in self._listener:
            self._client._on_error(topic_constants.ERROR,
                                   event_constants.NOT_LISTENING,
                                   pattern)
            future = concurrent.Future()
            future.set_result(None)
            return future

        listener = self._listener[pattern]

        if not listener.destroy_pending:
            listener.send_destroy()
        else:
            self._ack_timeout_registry.add(pattern, actions.UNLISTEN)
            listener.destroy()
            del self._listener[pattern]

        return listener.send_future

    def handle(self, message):
        action = message['action']
        data = message['data']

        if action == actions.ACK:
            name = message['data'][1]
        else:
            name = message['data'][0]

        if action == actions.EVENT:
            if data and len(data) == 2:
                self._emitter.emit(
                    name, message_parser.convert_typed(data[1], self._client))
            else:
                self._emitter.emit(name)

            return

        if (action == actions.ACK and data[0] == actions.UNLISTEN and
                (name in self._listener) and
                self._listener[name].destroy_pending):
            self._listener[name].destroy()
            del self._listener[name]
            return
        elif name in self._listener:
            self._listener[name]._on_message(message)
            return
        elif action in (actions.SUBSCRIPTION_FOR_PATTERN_REMOVED,
                        actions.SUBSCRIPTION_HAS_PROVIDER):
            return

        if action == actions.ACK:
            self._ack_timeout_registry.clear(message)
            return

        if action == actions.ERROR:
            if data[0] == event_constants.MESSAGE_DENIED:
                self._ack_timeout_registry.remove(message['data'][1],
                                                  message['data'][2])

            elif data[0] == event_constants.NOT_SUBSCRIBED:
                self._ack_timeout_registry.remove(message['data'][1],
                                                  actions.UNSUBSCRIBE)

            message['processedError'] = True
            self._client._on_error(topic_constants.EVENT, data[0], data[1])
            return

        self._client._on_error(topic_constants.EVENT,
                               event_constants.UNSOLICITED_MESSAGE,
                               name)

    def _resubscribe(self):
        for event in self._emitter._events:
            self._connection.send_message(topic_constants.EVENT,
                                          actions.SUBSCRIBE,
                                          [event])



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.constants import message as message_constants
from deepstreampy.constants import topic, event, actions, types

import json
import sys


def parse(raw_messages, client):
    parsed_messages = []
    raw_messages = raw_messages.split(message_constants.MESSAGE_SEPERATOR)
    for msg in raw_messages:
        # Ensure msg is not an empty string
        if msg:
            parsed_messages.append(_parse_message(msg, client))
    return parsed_messages


def _parse_message(message, client):
    parts = message.split(message_constants.MESSAGE_PART_SEPERATOR)
    if len(parts) < 2:
        # A valid message consists of at least 2 parts: action and topic
        client._on_error(topic.ERROR,
                         event.MESSAGE_PARSE_ERROR,
                         'Insufficient message parts')
        return

    if not actions.reverse_lookup(parts[1]):
        client._on_error(topic.ERROR, event.MESSAGE_PARSE_ERROR,
                         'Unknown action {0}'.format(parts[1]))
        return

    return {'raw': message,
            'topic': parts[0],
            'action': parts[1],
            'data': parts[2:]}


def convert_typed(value, client):
    value_type = value[0]

    if value_type == types.TRUE:
        return True

    if value_type == types.FALSE:
        return False

    if value_type == types.STRING:
        return value[1:]

    if value_type == types.NULL:
        return None

    if value_type == types.OBJECT:
        try:
            return json.loads(value[1:])
        except ValueError as e:
            client._on_error(topic.ERROR, event.MESSAGE_PARSE_ERROR, str(e))
            return

    if value_type == types.NUMBER:
        if sys.version_info < (3,):
            num_types = (int, long, float, complex)
        else:
            num_types = (int, float, complex)

        for num_type in num_types:
            try:
                return num_type(value[1:])
            except ValueError:
                pass

    client._on_error(topic.ERROR, event.MESSAGE_PARSE_ERROR,
                     'UNKNOWN_TYPE ({0})'.format(value))



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.message import connection
from deepstreampy import constants
from deepstreampy.record import RecordHandler
from deepstreampy.event import EventHandler
from deepstreampy.rpc import RPCHandler
from deepstreampy.presence import PresenceHandler

from pyee import EventEmitter
from tornado import gen


@gen.coroutine
def connect(url, **options):
    """Establish a connection to a deepstream.io server.

    Args:
        url (string): URL to connect to
        **options: The options for the client

    Returns:
        tornado.concurrent.Future: A future that resolves with an instance of
            ``deepstreampy.client.Client`` when a connection is established. If
            the client is unable to connect, the future will have the
            appropriate exception set.
    """
    client = Client(url, **options)
    yield client.connect()
    raise gen.Return(client)


class Client(EventEmitter):
    """
    deepstream.io Python client based on tornado.
    """

    def __init__(self, url, **options):
        """Creates the client but doesn't connect to the server.

        Args:
            url (str): The url to connect to
            options
        """
        super(Client, self).__init__()
        self._connection = connection.Connection(self, url, **options)
        self._presence = PresenceHandler(self._connection, self, **options)
        self._event = EventHandler(self._connection, self, **options)
        self._rpc = RPCHandler(self._connection, self, **options)
        self._record = RecordHandler(self._connection, self, **options)
        self._message_callbacks = dict()

        self._message_callbacks[
            constants.topic.PRESENCE] = self._presence.handle

        self._message_callbacks[
            constants.topic.EVENT] = self._event.handle

        self._message_callbacks[
            constants.topic.RPC] = self._rpc.handle

        self._message_callbacks[
            constants.topic.RECORD] = self._record.handle

        self._message_callbacks[constants.topic.ERROR] = self._on_error

    def connect(self, callback=None):
        """Establishes a connection to the url given to the constructor.

        Args:
            callback (callable): Will be called when connection is established
                without any arguments

        Returns:
            tornado.concurrent.Future: A future that resolves when the
                connection is established, or raises an exception if the client
                is unable to connect
        """
        return self._connection.connect(callback)

    def close(self):
        self._connection.close()

    def login(self, auth_params):
        """Sends authentication parameters to the server.

        If the connection is not yet established the authentication parameter
        will be stored and send once it becomes available
        Can be called multiple times until either the connection is
        authenticated or the max auth attempts is reached.

        Args:
            auth_params: JSON serializable data structure, it's up to the
                permission handler on the server to make sense of them.
            callback (callable): Will be called with True in case of success, or
                False, error_type, error_message in case of failure
        """
        return self._connection.authenticate(auth_params)

    def _on_message(self, message):
        if message['topic'] in self._message_callbacks:
            self._message_callbacks[message['topic']](message)
            return
        else:
            self._on_error(message['topic'],
                           constants.event.MESSAGE_PARSE_ERROR,
                           ('Received message for unknown topic ' +
                            message['topic']))
            return

        if message['action'] == constants.actions.ERROR:
            self._on_error(message['topic'],
                           message['action'],
                           message['data'][0] if len(message['data']) else None)

    def _on_error(self, topic, event, msg=None):
        if event in (constants.event.ACK_TIMEOUT,
                     constants.event.RESPONSE_TIMEOUT):
            if (self._connection.state ==
                    constants.connection_state.AWAITING_AUTHENTICATION):
                error_msg = ('Your message timed out because you\'re not '
                             'authenticated. Have you called login()?')
                self._connection.io_loop.call_later(
                    0.1, lambda: self._on_error(event.NOT_AUTHENTICATED,
                                                constants.topic.ERROR,
                                                error_msg))

        if self.listeners('error'):
            self.emit('error', msg, event, topic)
            self.emit(event, topic, msg)
        else:
            raw_error_message = event + ': ' + msg

            if topic:
                raw_error_message += ' (' + topic + ')'

            raise ValueError(raw_error_message)

    @property
    def connection_state(self):
        return self._connection.state

    @property
    def record(self):
        return self._record

    @property
    def event(self):
        return self._event

    @property
    def rpc(self):
        return self._rpc

    @property
    def presence(self):
        return self._presence

    @property
    def io_loop(self):
        return self._connection.io_loop



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.constants import topic as topic_constants
from deepstreampy.constants import actions as action_constants
from deepstreampy.constants import event as event_constants
from deepstreampy.constants import connection_state
from deepstreampy.message import message_parser, message_builder
from deepstreampy.utils import ResubscribeNotifier, SingleNotifier, Listener
from deepstreampy.utils import str_types
from deepstreampy.constants import merge_strategies
from deepstreampy import jsonpath

from pyee import EventEmitter
from tornado import gen, concurrent

import json
from functools import partial
from copy import deepcopy

ALL_EVENT = 'ALL_EVENT'
ENTRY_ADDED_EVENT = 'ENTRY_ADDED_EVENT'
ENTRY_REMOVED_EVENT = 'ENTRY_REMOVED_EVENT'
ENTRY_MOVED_EVENT = 'ENTRY_MOVED_EVENT'


class Record(EventEmitter, object):
    def __init__(self, name, connection, options, client):
        super(Record, self).__init__()
        self.name = name
        self.usages = 0
        self._connection = connection
        self._client = client
        self._options = options

        self._has_provider = False
        self._is_ready = False
        self._is_destroyed = False
        self._data = {}
        self._version = None
        self._old_value = None
        self._old_path_values = None
        self._queued_method_calls = list()
        self._write_callbacks = {}
        self.merge_strategy = merge_strategies.remote_wins

        self._emitter = EventEmitter()

        if 'merge_strategy' in options:
            self.merge_strategy = options['merge_strategy']

        self._resubscribe_notifier = ResubscribeNotifier(
            client, self._send_read)
        record_read_ack_timeout = options.get("recordReadAckTimeout", 15)
        self._read_ack_timeout = client.io_loop.call_later(
            record_read_ack_timeout,
            partial(self._on_timeout, event_constants.ACK_TIMEOUT))

        record_read_timeout = options.get("recordReadTimeout", 15)
        self._read_timeout = client.io_loop.call_later(
            record_read_timeout,
            partial(self._on_timeout, event_constants.RESPONSE_TIMEOUT))

        self._record_delete_timeout = options.get("recordDeleteTimeout", 15)

        self._delete_ack_timeout = None
        self._discard_timeout = None

    def get(self, path=None):
        """
        Returns a copy of either the entire dataset of the record or, if called
        with a path - the value of that path within the record's dataset.

        Returning a copy rather than the actual value helps to prevent the
        record getting out of sync due to unintentional changes to its data.

        Args:
            path (str, optional): a JSON path
        """
        return self._get_path(path)

    def set(self, data, path=None, callback=None):
        """
        Sets the value of either the entire dataset or of a specific path within
        the record and submits the changes to the server.

        If the new data is equal to the current data, nothing happens.

        Args:
            data: the new value of the data
            path (str, optional): a JSON path
            callback (callable)
        """
        config = {}
        if callback:
            state = self._client.connection_state
            if state in (connection_state.CLOSED,
                         connection_state.RECONNECTING):

                callback('Connection error: error updating record as '
                         'connection was closed')
                return
            else:
                config['writeSuccess'] = True
                self._set_up_callback(self.version, callback)

        if path is None and not isinstance(data, (dict, list)):
            raise ValueError(
                "Invalid record data {0}: Record data must be a dict or list.")

        if self._check_destroyed('set'):
            return

        if not self._is_ready:
            self._queued_method_calls.append(partial(self.set, data, path))
            return

        old_value = self._data
        deep_copy = self._options.get('recordDeepCopy', True)
        new_value = jsonpath.set(old_value, path, data, deep_copy)

        if new_value == old_value:
            if callback:
                callback(None)
            return

        self._send_update(path, data, config)
        self._apply_change(new_value)

    def subscribe(self, callback, path=None, trigger_now=False):
        """
        Subscribe to changes to the record's dataset.

        When called with a path it will only subscribe to updates to that path,
        rather than the entire record.

        Args:
            callback (callable)
            path (str, optional): a JSON path to subscribe for
            trigger_now (bool): specifies whether the callback should be invoked
                immediately with the current value
        """
        if self._check_destroyed('subscribe'):
            return

        if path is None:
            event = ALL_EVENT
        else:
            event = path

        self._emitter.on(event, callback)

        if trigger_now and self._is_ready:
            if path:
                callback(jsonpath.get(self._data, path, True))
            else:
                callback(self._data)

    def unsubscribe(self, callback, path=None):
        """
        Remove a subscription that was previously made.

        Args:
            callback (callable)
            path (str, optional): the JSON path to unsibscribe for
        """
        if self._check_destroyed('unsubscribe'):
            return

        if path is None:
            event = ALL_EVENT
        else:
            event = path

        self._emitter.remove_listener(event, callback)

    def discard(self):
        """
        Remove all change listeners and notify the server that the client is no
        longer interested in updates for this record.
        """
        future = concurrent.Future()

        if self._check_destroyed('discard'):
            return

        def ready_callback(record):
            self.usages -= 1

            if self.usages <= 0:
                self.emit('destroyPending')
                self._discard_timeout = self._client.io_loop.call_later(
                    1, partial(self._on_timeout, event_constants.ACK_TIMEOUT))

                send_future = self._connection.send_message(
                    topic_constants.RECORD, action_constants.UNSUBSCRIBE,
                    [self.name])
                send_future.add_done_callback(
                    lambda f: future.set_result(f.result()))

        self.when_ready(ready_callback)
        return future

    def delete(self):
        """
        Delete the record on the server.
        """
        future = concurrent.Future()

        if self._check_destroyed('delete'):
            return

        def ready_callback(record):
            self.emit('destroyPending')
            self._delete_ack_timeout = self._client.io_loop.call_later(
                self._record_delete_timeout,
                partial(self._on_timeout, event_constants.DELETE_TIMEOUT))

            send_future = self._connection.send_message(
                topic_constants.RECORD, action_constants.DELETE, [self.name])
            send_future.add_done_callback(
                lambda f: future.set_result(f.result()))

        self.when_ready(ready_callback)
        return future

    def when_ready(self, callback):
        if self._is_ready:
            callback(self)
        else:
            self.once('ready', partial(callback, self))

    def _set_up_callback(self, current_version, callback):
        new_version = (current_version or 0) + 1
        self._write_callbacks[new_version] = callback

    def _on_message(self, message):
        action = message['action']

        if action == action_constants.READ:
            if self.version is None:
                self._client.io_loop.remove_timeout(self._read_timeout)
                self._on_read(message)
            else:
                self._apply_update(message)

        elif action == action_constants.ACK:
            self._process_ack_message(message)

        elif action in (action_constants.UPDATE, action_constants.PATCH):
            self._apply_update(message)

        elif action == action_constants.WRITE_ACKNOWLEDGEMENT:
            versions = json.loads(message['data'][1])
            for version in versions:
                if version in self._write_callbacks:
                    callback = self._write_callbacks[version]
                    callback(
                        message_parser.convert_typed(message['data'][2],
                                                     self._client))
                    del self._write_callbacks[version]

        elif message['data'][0] == event_constants.VERSION_EXISTS:
            self._recover_record(message['data'][2],
                                 json.loads(message['data'][3]), message)

        elif action == event_constants.MESSAGE_DENIED:
            self._clear_timeouts()

        elif action == action_constants.SUBSCRIPTION_HAS_PROVIDER:
            has_provider = message_parser.convert_typed(
                message['data'][1], self._client)
            self._has_provider = has_provider
            self.emit('hasProviderChanged', has_provider)

    def _recover_record(self, remote_version, remote_data, message):
        if self.merge_strategy:
            self.merge_strategy(self, remote_data, remote_version,
                                partial(self._on_record_recovered,
                                        remote_version, remote_data, message))
        else:
            self.emit('error', event_constants.VERSION_EXISTS,
                      'received update for {0} but version is {1}'.format(
                          remote_version, self.version))

    def _on_record_recovered(self, remote_version, remote_data, message, error,
                             data):
        if not error:
            old_version = self.version
            self._version = int(remote_version)

            old_value = self._data

            new_value = jsonpath.set(old_value, None, data, True)

            if data == remote_data:
                self._apply_change(data)

                callback = self._write_callbacks.get(self.version, None)
                if callback:
                    callback(None)
                    if remote_version in self._write_callbacks.keys():
                        del self._write_callbacks[remote_version]

                return

            config = message['data'][4] if len(message['data']) >= 5 else None
            if config and json.loads(config)['writeSuccess']:
                callback = self._write_callbacks[old_version]
                del self._write_callbacks[old_version]
                self._set_up_callback(self.version, callback)

            self._send_update(None, data, config)
            self._apply_change(new_value)
        else:
            self.emit('error', event_constants.VERSION_EXISTS,
                      'received update for {0} but version is {1}'.format(
                          remote_version, self.version))

    def _process_ack_message(self, message):
        acknowledge_action = message['data'][0]

        if acknowledge_action == action_constants.SUBSCRIBE:
            self._client.io_loop.remove_timeout(self._read_ack_timeout)

        elif acknowledge_action == action_constants.DELETE:
            self.emit('delete')
            self._destroy()

        elif acknowledge_action == action_constants.UNSUBSCRIBE:
            self.emit('discard')
            self._destroy()

    def _apply_update(self, message):
        version = int(message['data'][1])
        if message['action'] == action_constants.PATCH:
            data = message_parser.convert_typed(message['data'][3],
                                                self._client)
        else:
            data = json.loads(message['data'][2])

        if self.version is None:
            self._version = version
        elif self.version + 1 != version:
            if message['action'] == action_constants.PATCH:
                self._connection.send_message(topic_constants.RECORD,
                                              action_constants.SNAPSHOT,
                                              [self.name])
            else:
                self._recover_record(version, data, message)

            return

        self._begin_change()
        self._version = version
        if message['action'] == action_constants.PATCH:
            jsonpath.set(self._data, message['data'][2], data, False)
        else:
            self._data = data

        self._complete_change()

    def _send_update(self, path, data, config):
        self._version += 1
        if not path:
            if config:
                msg_data = [self.name, self.version, data, config]
            else:
                msg_data = [self.name, self.version, data]
            self._connection.send_message(topic_constants.RECORD,
                                          action_constants.UPDATE, msg_data)
        else:
            if config:
                msg_data = [
                    self.name, self.version, path,
                    message_builder.typed(data), config
                ]
            else:
                msg_data = [
                    self.name, self.version, path,
                    message_builder.typed(data)
                ]
            self._connection.send_message(topic_constants.RECORD,
                                          action_constants.PATCH, msg_data)

    def _apply_change(self, new_data):
        if self.is_destroyed:
            return

        old_data = self._data
        self._data = new_data

        if not self._emitter._events:
            return

        paths = self._emitter._events.keys()
        for path in paths:
            if path == 'new_listener':
                continue

            if path == 'ALL_EVENT' and new_data != old_data:
                self._emitter.emit(ALL_EVENT, new_data)
                continue

            new_value = jsonpath.get(new_data, path, False)
            old_value = jsonpath.get(old_data, path, False)

            if new_value != old_value:
                self._emitter.emit(path, self._get_path(path))

    def _on_read(self, message):
        self._begin_change()
        self._version = int(message['data'][1])
        self._data = json.loads(message['data'][2])
        self._complete_change()
        self._set_ready()

    def _set_ready(self):
        self._is_ready = True
        for call in self._queued_method_calls:
            call()
        self._queued_method_calls = []
        self.emit('ready')

    def _send_read(self):
        """
        Sends the read message, either initially at record creation or after a
        lost connection has been re-established.
        """
        return self._connection.send_message(
            topic_constants.RECORD, action_constants.CREATEORREAD, [self.name])

    def _get_path(self, path=None):
        deep_copy = self._options.get('recordDeepCopy', False)
        return jsonpath.get(self._data, path, deep_copy)

    def _begin_change(self):
        if not self._emitter._events:
            return

        # Hacky way of getting active listeners, except a special one
        paths = [
            event for event in self._emitter._events.keys()
            if event != 'new_listener'
        ]

        self._old_path_values = dict()

        if self._emitter.listeners(ALL_EVENT):
            self._old_value = deepcopy(self._data)

        for path in paths:
            if path != ALL_EVENT:
                self._old_path_values[path] = jsonpath.get(
                    self._data, path, True)

    def _complete_change(self):
        if (self._emitter.listeners(ALL_EVENT)
                and self._old_value != self._data):
            self._emitter.emit(ALL_EVENT, self._data)

        self._old_value = None

        if not self._old_path_values:
            return

        for path in self._old_path_values:
            current_value = self._get_path(
                path)  #jsonpath.get(self._data, path, True)

            if current_value != self._old_path_values[path]:
                self._emitter.emit(path, current_value)

        self._old_path_values = None

    def _clear_timeouts(self):
        if self._read_ack_timeout:
            self._client.io_loop.remove_timeout(self._read_ack_timeout)
        if self._discard_timeout:
            self._client.io_loop.remove_timeout(self._discard_timeout)
        if self._delete_ack_timeout:
            self._client.io_loop.remove_timeout(self._delete_ack_timeout)

    def _check_destroyed(self, method_name):
        if self._is_destroyed:
            self.emit(
                'error',
                "Can't invoke {0}. Record {1} is already destroyed".format(
                    method_name, self.name))
            return True

        return False

    def _on_timeout(self, timeout_type):
        self._clear_timeouts()
        self.emit('error', timeout_type)

    def _destroy(self):
        self._clear_timeouts()
        self._emitter.remove_all_listeners()
        self._resubscribe_notifier.destroy()
        self._is_destroyed = True
        self._is_ready = False
        self._client = None
        self._connection = None

    @property
    def has_provider(self):
        return self._has_provider

    @property
    def is_destroyed(self):
        return self._is_destroyed

    @property
    def is_ready(self):
        return self._is_ready

    @property
    def version(self):
        return self._version


class List(Record):
    def __init__(self, name, connection, options, client):
        super(List, self).__init__(name, connection, options, client)
        self._before_structure = None
        self._has_add_listener = None
        self._has_remove_listener = None
        self._has_move_listener = None

    def get(self):
        """
        Return the list of entries or an empty array if the list hasn't been
        populated yet.
        """
        entries = super(List, self).get()

        if not isinstance(entries, list):
            return []

        return entries

    def set(self, entries):
        """
        Update the list with a new set of entries.

        Args:
            entries (list): the new entries
        """
        error_msg = 'entries must be a list of record names'

        if not isinstance(entries, list):
            raise ValueError(error_msg)

        for entry in entries:
            if not isinstance(entry, str_types):
                raise ValueError(error_msg)

        if not self.is_ready:
            self._queued_method_calls.append(partial(self.set, entries))
        else:
            self._before_change()
            super(List, self).set(entries)
            self._after_change()

    def remove_entry(self, entry):
        """
        Remove the entry from the list.

        Args:
            entry (str): the entry to remove
        """
        if not self.is_ready:
            self._queued_method_calls.append(partial(self.remove_entry, entry))

        current_entries = deepcopy(super(List, self).get())
        current_entries.remove(entry)

        self.set(current_entries)

    def remove_at(self, index):
        """
        Remove the entry at the specified index.

        Args:
            index (int): the index of the entry to remove
        """
        if not self.is_ready:
            self._queued_method_calls.append(partial(self.remove_at, index))

        current_entries = deepcopy(super(List, self).get())
        del current_entries[index]
        self.set(current_entries)

    def add_entry(self, entry, index=None):
        """
        Add an entry to the list.

        Args:
            entry (str): the entry to add
            index (int): the index at which to add the entry
        """
        if not self.is_ready:
            self._queued_method_calls.append(
                partial(self.add_entry, entry, index))

        entries = deepcopy(self.get())
        if index is not None:
            entries.insert(index, entry)
        else:
            entries.append(entry)

        self.set(entries)

    def subscribe(self, callback):
        """
        Proxies the underlying Record's subscribe method.
        """
        super(List, self).subscribe(callback)

    def unsubscribe(self, callback):
        """
        Proxies the underlying Record's unsubscribe method.
        """
        super(List, self).unsubscribe(callback)

    def _apply_update(self, message):
        if message['action'] == action_constants.PATCH:
            raise ValueError('PATCH is not supported for Lists')

        if message['data'][2][0] != '[':
            message['data'][2] = '[]'

        self._before_change()
        super(List, self)._apply_update(message)
        self._after_change()

    def _before_change(self):
        self._has_add_listener = len(self.listeners(ENTRY_ADDED_EVENT)) > 0
        self._has_remove_listener = len(
            self.listeners(ENTRY_REMOVED_EVENT)) > 0
        self._has_move_listener = len(self.listeners(ENTRY_MOVED_EVENT)) > 0

        if (self._has_add_listener or self._has_remove_listener
                or self._has_move_listener):
            self._before_structure = self._get_structure()
        else:
            self._before_structure = None

    def _after_change(self):
        if self._before_structure is None:
            return

        after = self._get_structure()
        before = self._before_structure

        if self._has_remove_listener:
            self._after_change_remove_listener(before, after)

        if self._has_add_listener or self._has_move_listener:
            self._after_change_move_add_listener(before, after)

    def _after_change_remove_listener(self, before, after):
        for entry in before:
            if (entry not in after or len(after[entry]) < len(before[entry])):
                for n in before[entry]:
                    if entry not in after or n not in after[entry]:
                        self.emit(ENTRY_REMOVED_EVENT, entry, n)

    def _after_change_move_add_listener(self, before, after):
        for entry in after:
            if entry not in before:
                for n in after[entry]:
                    self.emit(ENTRY_ADDED_EVENT, entry, n)
            elif before[entry] != after[entry]:
                added = len(before[entry]) != len(after[entry])
                for n in after[entry]:
                    if added and n not in before[entry]:
                        self.emit(ENTRY_ADDED_EVENT, entry, n)
                    elif not added:
                        self.emit(ENTRY_MOVED_EVENT, entry, n)

    def _get_structure(self):
        structure = {}
        entries = super(List, self).get()

        for i, entry in enumerate(entries):
            if entry in structure:
                structure[entry].append(i)
            else:
                structure[entry] = [i]

        return structure

    @property
    def is_empty(self):
        return len(self.get()) == 0


class RecordHandler(EventEmitter, object):
    def __init__(self, connection, client, **options):
        super(RecordHandler, self).__init__()
        self._options = options
        self._connection = connection
        self._client = client
        self._records = {}
        self._lists = {}
        self._listeners = {}
        self._destroy_emitter = EventEmitter()

        record_read_timeout = options.get("recordReadTimeout", 15)

        self._has_registry = SingleNotifier(
            client, connection, topic_constants.RECORD, action_constants.HAS,
            record_read_timeout)

        self._snapshot_registry = SingleNotifier(
            client, connection, topic_constants.RECORD,
            action_constants.SNAPSHOT, record_read_timeout)

    @gen.coroutine
    def get_record(self, name):
        """
        Return an existing record or create a new one.

        Args:
            name (str): the unique name of the record
        """
        if name in self._records:
            record = self._records[name]
        else:
            record = Record(name, self._connection, self._options,
                            self._client)
            record.on('error', partial(self._on_record_error, name))
            record.on('destroyPending', partial(self._on_destroy_pending,
                                                name))
            record.on('delete', partial(self._remove_record, name))
            record.on('discard', partial(self._remove_record, name))

            self._records[name] = record

            record.usages += 1

        yield record._send_read()
        raise gen.Return(record)

    @gen.coroutine
    def get_list(self, name):
        """
        Return an exising list or create a new one.

        Args:
            name (str): the unique name of the list
            list_options (dict): a dict of parameters for this particular list
        """
        if name in self._lists:
            _list = self._lists[name]
        else:
            _list = List(name, self._connection, self._options, self._client)

            self._lists[name] = _list

        if name not in self._records:
            self._records[name] = _list

            _list.on('error', partial(self._on_record_error, name))
            _list.on('destroyPending', partial(self._on_destroy_pending, name))
            _list.on('delete', partial(self._remove_record, name))
            _list.on('discard', partial(self._remove_record, name))

        self._records[name].usages += 1

        yield _list._send_read()
        raise gen.Return(_list)

    def get_anonymous_record(self):
        """
        Return an anonymous record.
        """
        future = concurrent.Future()
        future.set_result(AnonymousRecord(self))
        return future

    def listen(self, pattern, callback):
        """
        Listen for record subscriptions made by this or other clients. This is
        useful to create "active" data providers, e.g. providers that only
        provide data for a particular record if a user is actually interested in
        it.

        Args:
            pattern (str): A combination of alpha numeric characters and
                wildcards(*)
            callback (callable):
        """
        if pattern in self._listeners:
            self._client._on_error(topic_constants.RECORD,
                                   event_constants.LISTENER_EXISTS, pattern)
            future = concurrent.Future()
            future.set_result(None)
        else:
            listener = Listener(topic_constants.RECORD, pattern, callback,
                                self._options, self._client, self._connection)
            self._listeners[pattern] = listener
            future = listener.send_future

        return future

    def unlisten(self, pattern):
        """
        Remove a listener that was previously registered with `listen`.

        Args:
            pattern (str): A combination of alpha numeric characters and
                wildcards(*)
        """
        if pattern in self._listeners:
            listener = self._listeners[pattern]
            if not listener.destroy_pending:
                listener.send_destroy()
                future = concurrent.Future()
                future.set_result(None)
            else:
                future = listener.destroy()
                del self._listeners[pattern]
        else:
            self._client._on_error(topic_constants.RECORD,
                                   event_constants.NOT_LISTENING, pattern)
            future = concurrent.Future()
            future.set_result(None)

        return future

    def snapshot(self, name, callback):
        """
        Retrieve the current record data without subscribing to changes.

        Args:
            name (str): the unique name of the record
            callback (callable):
        """
        if name in self._records and self._records[name].is_ready:
            callback(None, self._records[name].get())
            future = concurrent.Future()
            future.set_result(None)
        else:
            future = self._snapshot_registry.request(name, callback)

        return future

    def has(self, name, callback):
        """
        Check whether the record exists.

        Args:
            name (str): the unique name of the record
            callback (callable):
        """
        if name in self._records:
            callback(None, True)
            future = concurrent.Future()
            future.set_result(None)
        else:
            future = self._has_registry.request(name, callback)

        return future

    def _process_message(self, message, name):
        action = message['action']
        data = message['data']
        processed = False

        if (action == action_constants.READ
                and self._snapshot_registry.has_request(name)):
            processed = True
            self._snapshot_registry.receive(name, None, json.loads(data[2]))

        if (action == action_constants.HAS
                and self._has_registry.has_request(name)):
            processed = True
            record_exists = message_parser.convert_typed(data[1], self._client)
            self._has_registry.receive(name, None, record_exists)

        listener = self._listeners.get(name, None)
        if (action == action_constants.ACK
                and data[0] == action_constants.UNLISTEN and listener
                and listener.destroy_pending):
            processed = True
            listener.destroy()
            del self._listeners[name]
            del listener
        elif listener:
            processed = True
            listener._on_message(message)
        elif action in (action_constants.SUBSCRIPTION_FOR_PATTERN_REMOVED,
                        action_constants.SUBSCRIPTION_HAS_PROVIDER):
            processed = True

        return processed

    def handle(self, message):
        action = message['action']
        data = message['data']

        if (action == action_constants.ERROR
                and data[0] not in (event_constants.VERSION_EXISTS,
                                    action_constants.SNAPSHOT,
                                    action_constants.HAS)):
            message['processedError'] = True
            self._client._on_error(topic_constants.RECORD, message['data'][0],
                                   message['data'][1])
            return

        if action in (action_constants.ACK, action_constants.ERROR):
            name = data[1]

            if data[0] in (action_constants.DELETE,
                           action_constants.UNSUBSCRIBE):
                name = message['data'][1]
                self._destroy_emitter.emit('destroy_ack_' + name, message)

                if (message['data'][0] == action_constants.DELETE
                        and name in self._records):
                    self._records[name]._on_message(message)
                return

            if data[0] in (action_constants.SNAPSHOT, action_constants.HAS):
                message['processedError'] = True
                error = message['data'][2]
                self._snapshot_registry.receive(name, error, None)
                return
        else:
            name = message['data'][0]

        processed = False

        if name in self._records:
            processed = True
            self._records[name]._on_message(message)

        processed = self._process_message(message, name) or processed

        if not processed:
            self._client._on_error(topic_constants.RECORD,
                                   event_constants.UNSOLICITED_MESSAGE, name)

    def _on_record_error(self, record_name, error):
        message = "No ACK message received in time for {}".format(record_name)
        self._client._on_error(topic_constants.RECORD, error, message)

    def _on_destroy_pending(self, record_name):
        on_message = self._records[record_name]._on_message
        self._destroy_emitter.once('destroy_ack_' + record_name, on_message)
        self._remove_record(record_name)

    def _remove_record(self, record_name):
        if record_name in self._records:
            del self._records[record_name]
        elif record_name in self._lists:
            del self._lists[record_name]


class AnonymousRecord(EventEmitter, object):
    def __init__(self, record_handler):
        super(AnonymousRecord, self).__init__()
        self._record_handler = record_handler
        self._name = None
        self._record = None
        self._subscriptions = []
        self._proxy_method('delete')
        self._proxy_method('set')
        self._proxy_method('discard')

    def get(self, path=None):
        """
        Proxies the actual record's get method.

        Args:
            path (str, optional): a JSON path. If not provided, the entire
                record is returned
        """
        if self._record is None:
            return None

        return self._record.get(path)

    def subscribe(self, callback, path=None):
        """
        Proxies the actual record's subscribe method.

        Args:
            callback (callable):
            path (str): a JSON path. If not provided, the subscription is for
                the entire record.
        """
        self._subscriptions.append((callback, path))

        if self._record is not None:
            self._record.subscribe(callback, path, True)

    def unsubscribe(self, callback, path=None):
        """
        Proxies the actual record's unsubscribe method.

        Args:
            callback (callable):
            path (str): a JSON path. If not provided, the subscription is for
                the entire record.
        """
        self._subscriptions.remove((callback, path))

        if self._record is not None:
            self._record.unsubscribe(callback, path)

    def _on_record_get(self, record):
        self._record = record

    @property
    def name(self):
        return self._name

    @name.setter
    @gen.coroutine
    def name(self, value):
        self._name = value

        if self._record is not None and not self._record.is_destroyed:
            for subscription in self._subscriptions:
                self._record.unsubscribe(*subscription)

            self._record.discard()

        record_future = self._record_handler.get_record(value)

        self._record = yield record_future

        for subscription in self._subscriptions:
            self._record.subscribe(*subscription, trigger_now=True)

        self._record.when_ready(partial(self.emit, "ready"))
        self.emit("nameChanged", value)

    def _proxy_method(self, method_name):
        method = partial(self._call_method_on_record, method_name)
        setattr(self, method_name, method)

    def _call_method_on_record(self, method_name, *args, **kwargs):
        if self._record is None:
            raise AttributeError(
                "Can't invoke {}. AnonymousRecord not initialised. "
                "Set `name` first.")

        getattr(self._record, method_name)(*args, **kwargs)



"""Deepstream RPC handling."""

from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.constants import topic as topic_constants
from deepstreampy.constants import actions
from deepstreampy.constants import event as event_constants
from deepstreampy.message import message_builder
from deepstreampy.message import message_parser
from deepstreampy import utils

from tornado import concurrent
from tornado import gen

from functools import partial


class RPCResponse(object):
    """Allows a RPC provider to respond to a request.

    Attributes:
        auto_ack (bool): Specifies whether requests should be auto acknowledged
    """

    def __init__(self, connection, name, correlation_id):
        """
        Args:
            connection (deepstreampy.client._Connection): The current connection
            name (str): The name of the RPC
            correlation_id (str): Correlation ID of the RPC
        """
        self._connection = connection
        self._name = name
        self._correletaion_id = correlation_id
        self._is_acknowledged = False
        self._is_complete = False
        self.auto_ack = True

        self._connection._io_loop.add_callback(self._perform_auto_ack)

    def ack(self):
        """Acknowledge the receiving the request.

        Will happen implicitly unless the request callback sets ``auto_ack``
        to False.

        """
        if not self._is_acknowledged:
            future = self._connection.send_message(
                topic_constants.RPC,
                actions.ACK,
                [actions.REQUEST, self._name, self._correletaion_id])
            self._is_acknowledged = True
        else:
            future = concurrent.Future()
            future.set_result(None)

        return future

    def reject(self):
        """Reject the request.

        This might be necessary if the client is already processing a large
        number of requests. If deepstream receives a rejection message it will
        try to route the request to another provider - or return a
        NO_RPC_PROVIDER error if there are no providers left.

        If autoAck is disabled and the response is sent before the ack message
        the request will still be completed and the ack message ignored.

        """
        self.auto_ack = False
        self._is_complete = True
        self._is_acknowledged = True
        return self._connection.send_message(
            topic_constants.RPC,
            actions.REJECTION,
            [self._name, self._correletaion_id])

    def send(self, data):
        """Complete the request by sending the response data to the server.

        Args:
            data: JSON serializable data to send to the server.
        """
        if self._is_complete:
            raise ValueError('RPC {0} already completed'.format(self._name))
        self.ack()

        typed_data = message_builder.typed(data)
        self._is_complete = True

        return self._connection.send_message(
            topic_constants.RPC,
            actions.RESPONSE,
            [self._name, self._correletaion_id, typed_data])

    def error(self, error_str):
        """Notify the server that an error has occured.

        This will also complete the RPC.
        """
        self.auto_ack = False
        self._is_complete = True
        self._is_acknowledged = True
        return self._connection.send_message(
            topic_constants.RPC,
            actions.ERROR,
            [error_str, self._name, self._correletaion_id])

    def _perform_auto_ack(self):
        if self.auto_ack:
            self.ack()

class RPCException(Exception):

    def __init__(self, message):
        super(RPCException, self).__init__(message)

class RPC(object):
    """Represents a single RPC made from the client to the server.

    Encapsulates logic around timeouts and converts the incoming response data.
    """

    def __init__(self, future, client, **options):
        self._options = options
        self._future = future
        self._client = client
        self._connection = client._connection

        self._ack_timeout = self._connection._io_loop.call_later(
            options.get('rpcAckTimeout', 6),
            partial(self.error, event_constants.ACK_TIMEOUT))

        self._response_timeout = self._connection._io_loop.call_later(
            options.get('rpcResponseTimeout', 6),
            partial(self.error, event_constants.RESPONSE_TIMEOUT))

    def ack(self):
        self._connection._io_loop.remove_timeout(self._ack_timeout)

    def respond(self, data):
        converted_data = message_parser.convert_typed(data, self._client)
        self._future.set_result(converted_data)
        self._complete()

    def error(self, error_msg):
        self._future.set_exception(RPCException(error_msg))
        self._complete()

    def _complete(self):
        self._connection._io_loop.remove_timeout(self._ack_timeout)
        self._connection._io_loop.remove_timeout(self._response_timeout)


class RPCHandler(object):

    def __init__(self, connection, client, **options):
        self._options = options
        self._connection = connection
        self._client = client
        self._rpcs = {}
        self._providers = {}
        self._provide_ack_timeouts = {}

        subscription_timeout = options.get("subscriptionTimeout", 15)
        self._ack_timeout_registry = utils.AckTimeoutRegistry(
            client, topic_constants.RPC, subscription_timeout)
        self._resubscribe_notifier = utils.ResubscribeNotifier(
            client, self._reprovide)

    def provide(self, name, callback):
        if not name:
            raise ValueError("invalid argument: name")
        if not callback:
            raise ValueError("invalid argument: callback")
        if not callable(callback):
            raise TypeError("expected callback to be a callable")
        if name in self._providers:
            raise ValueError("RPC {0} already registered".format(name))

        self._ack_timeout_registry.add(name, actions.SUBSCRIBE)
        self._providers[name] = callback

        return self._connection.send_message(topic_constants.RPC,
                                             actions.SUBSCRIBE,
                                             [name])

    def unprovide(self, name):
        if not name:
            raise ValueError("invalid argument name")

        if name in self._providers:
            del self._providers[name]
            self._ack_timeout_registry.add(name, actions.UNSUBSCRIBE)
            future = self._connection.send_message(topic_constants.RPC,
                                                   actions.UNSUBSCRIBE,
                                                   [name])
        else:
            future = concurrent.Future()
            future.set_result(None)

        return future

    @gen.coroutine
    def make(self, name, data):
        f = concurrent.Future()

        uid = utils.get_uid()
        typed_data = message_builder.typed(data)

        self._rpcs[uid] = RPC(f, self._client, **self._options)

        self._connection.send_message(
            topic_constants.RPC, actions.REQUEST, [name, uid, typed_data])

        result = yield f
        raise gen.Return(result)

    def _get_rpc(self, correlation_id, rpc_name, raw_message):
        if correlation_id not in self._rpcs:
            self._client._on_error(topic_constants.RPC,
                                   event_constants.UNSOLICITED_MESSAGE,
                                   raw_message)
            return

        rpc = self._rpcs[correlation_id]
        return rpc

    def _respond_to_rpc(self, message):
        name = message['data'][0]
        correlation_id = message['data'][1]

        if message['data'][2]:
            data = message_parser.convert_typed(message['data'][2],
                                                self._client)

        if name in self._providers:
            response = RPCResponse(self._connection, name, correlation_id)
            self._providers[name](data, response)
        else:
            self._connection.send_message(topic_constants.RPC,
                                          actions.REJECTION,
                                          [name, correlation_id])

    def handle(self, message):
        action = message['action']
        data = message['data']
        if action == actions.REQUEST:
            self._respond_to_rpc(message)
            return

        if (action == actions.ACK and
                (data[0] in (actions.SUBSCRIBE, actions.UNSUBSCRIBE))):
            self._ack_timeout_registry.clear(message)
            return

        if action == actions.ERROR:
            if data[0] in (event_constants.MESSAGE_PERMISSION_ERROR,
                           event_constants.INVALID_RPC_CORRELATION_ID):
                return

            if (data[0] == event_constants.MESSAGE_DENIED and
                    data[2] == actions.SUBSCRIBE):
                self._ack_timeout_registry.remove(data[1], actions.SUBSCRIBE)
                return

        if action in (actions.ERROR, actions.ACK):
            if (data[0] == event_constants.MESSAGE_DENIED and
                    data[2] == actions.REQUEST):
                correlation_id = data[3]
            else:
                correlation_id = data[2]
            rpc_name = data[1]
        else:
            rpc_name = data[0]
            correlation_id = data[1]

        rpc = self._get_rpc(correlation_id, rpc_name, message.get('raw', ''))
        if rpc is None:
            return

        if action == actions.ACK:
            rpc.ack()
        elif action == actions.RESPONSE:
            rpc.respond(data[2])
            del self._rpcs[correlation_id]
        elif action == actions.ERROR:
            message['processedError'] = True
            rpc.error(data[0])
            del self._rpcs[correlation_id]

    def _reprovide(self):
        for rpc_name in self._providers:
            self._connection.send_message(topic_constants.RPC,
                                          actions.SUBSCRIBE,
                                          [rpc_name])



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.utils import str_types
from deepstreampy import constants
from deepstreampy.message import message_builder, message_parser

from tornado import ioloop, concurrent, websocket, gen

from collections import deque
import errno
import time


class Connection(object):

    def __init__(self, client, url, **options):
        self._io_loop = ioloop.IOLoop.current()

        self._client = client
        self._original_url = url
        self._url = self._original_url
        self._websocket_handler = None

        self._auth_params = None
        self._auth_future = None
        self._connect_callback = None
        self._connect_error = None

        self._message_buffer = ""
        self._deliberate_close = False
        self._redirecting = False
        self._too_many_auth_attempts = False
        self._queued_messages = deque()
        self._reconnect_timeout = None
        self._reconnection_attempt = 0

        self._current_packet_message_count = 0
        self._send_next_packet_timeout = None
        self._last_heartbeat = None
        self._heartbeat_callback = None

        self._challenge_denied = False
        self._connection_auth_timeout = False

        self._current_message_reset_timeout = None
        self._state = constants.connection_state.CLOSED

        self._max_reconnect_attempts = options.get('maxReconnectAttempts', 3)
        self._reconnect_interval_increment = options.get(
            'reconnectAttemptInterval', 4)
        self._max_reconnect_interval = options.get(
            'maxReconnectInterval', 18)
        self._heartbeat_interval = options.get('heartbeatInterval', 100)

    def connect(self, callback=None):
        self._connect_callback = callback

        connect_future = websocket.websocket_connect(
                self._url,
                self._io_loop,
                callback=self._on_open,
                on_message_callback=self._on_data)

        return connect_future

    def _check_heartbeat(self):
        heartbeat_tolerance = self._heartbeat_interval * 2
        elapsed = time.time() - self._last_heartbeat
        if elapsed >= heartbeat_tolerance:
            self._io_loop.remove_timeout(self._heartbeat_callback)
            self._websocket_handler.close()
            self._on_error("heartbeat not received in the last 1000 milliseconds")
        else:
            self._heartbeat_callback = self._io_loop.call_later(
                self._heartbeat_interval, self._check_heartbeat)

    def _on_open(self, f):
        exception = f.exception()
        if exception:
            self._connect_error = exception
            self._on_error(self._connect_error)

            self._try_reconnect()

            return

        self._last_heartbeat = time.time()
        self._heartbeat_callback = self._io_loop.call_later(
            self._heartbeat_interval, self._check_heartbeat)

        self._websocket_handler = f.result()
        self._set_state(constants.connection_state.AWAITING_CONNECTION)

        if self._connect_callback:
            self._connect_callback()

    def _on_error(self, error):
        if self._heartbeat_callback:
            self._io_loop.remove_timeout(self._heartbeat_callback)
        self._set_state(constants.connection_state.ERROR)

        if isinstance(error, str_types):
            msg = error
        elif error.errno in (errno.ECONNRESET, errno.ECONNREFUSED):
            msg = ("Can't connect! Deepstream server unreachable on " +
                   self._url)
        else:
            msg = str(error)

        self._client._on_error(constants.topic.CONNECTION,
                               constants.event.CONNECTION_ERROR, msg)

    def start(self):
        self._io_loop.start()

    def stop(self):
        self._io_loop.stop()

    def close(self):
        if self._heartbeat_callback:
            self._io_loop.remove_timeout(self._heartbeat_callback)
        self._deliberate_close = True
        if self._websocket_handler:
            self._websocket_handler.close()

    def authenticate(self, auth_params):
        self._auth_params = auth_params
        self._auth_future = concurrent.Future()

        if (self._too_many_auth_attempts or
                self._challenge_denied or
                self._connection_auth_timeout):
            msg = "this client's connection was closed"
            self._client._on_error(constants.topic.ERROR,
                                   constants.event.IS_CLOSED,
                                   msg)
            self._auth_future.set_result(
                    {'success': False,
                     'error': constants.event.IS_CLOSED,
                     'message': msg})

        elif (self._deliberate_close and
              self._state == constants.connection_state.CLOSED):
            self.connect()
            self._deliberate_close = False
            self._client.once(constants.event.CONNECTION_STATE_CHANGED,
                              lambda: self.authenticate(auth_params))

        if self._state == constants.connection_state.AWAITING_AUTHENTICATION:
            self._send_auth_params()

        return self._auth_future

    def _send_auth_params(self):
        self._set_state(constants.connection_state.AUTHENTICATING)
        raw_auth_message = message_builder.get_message(
            constants.topic.AUTH,
            constants.actions.REQUEST,
            [self._auth_params])
        self._websocket_handler.write_message(raw_auth_message.encode())

    def _handle_auth_response(self, message):
        message_data = message['data']
        message_action = message['action']
        data_size = len(message_data)
        if message_action == constants.actions.ERROR:
            if (message_data and
                    message_data[0] == constants.event.TOO_MANY_AUTH_ATTEMPTS):
                self._deliberate_close = True
                self._too_many_auth_attempts = True
            else:
                self._set_state(
                    constants.connection_state.AWAITING_AUTHENTICATION)

            auth_data = (self._get_auth_data(message_data[1]) if
                         data_size > 1 else None)

            if self._auth_future:
                self._auth_future.set_result(
                    {'success': False,
                     'error': message_data[0] if data_size else None,
                     'message': auth_data})

        elif message_action == constants.actions.ACK:
            self._set_state(constants.connection_state.OPEN)

            auth_data = (self._get_auth_data(message_data[0]) if
                         data_size else None)

            # Resolve auth future and callback
            if self._auth_future:
                self._auth_future.set_result(
                    {'success': True, 'error': None, 'message': auth_data})

            self._send_queued_messages()

    def _handle_connection_response(self, message):
        action = message['action']
        data = message['data']
        if action == constants.actions.PING:
            self._last_heartbeat = time.time()
            ping_response = message_builder.get_message(
                constants.topic.CONNECTION, constants.actions.PONG)
            self.send(ping_response)
        elif action == constants.actions.ACK:
            self._set_state(constants.connection_state.AWAITING_AUTHENTICATION)
            if self._auth_params is not None:
                self._send_auth_params()
        elif action == constants.actions.CHALLENGE:
            challenge_response = message_builder.get_message(
                constants.topic.CONNECTION,
                constants.actions.CHALLENGE_RESPONSE,
                [self._url])
            self._set_state(constants.connection_state.CHALLENGING)
            self.send(challenge_response)
        elif action == constants.actions.REJECTION:
            self._challenge_denied = True
            self.close()
        elif action == constants.actions.REDIRECT:
            self._url = data[0]
            self._redirecting = True
            self.close()
        elif action == constants.actions.ERROR:
            if data[0] == constants.event.CONNECTION_AUTHENTICATION_TIMEOUT:
                self._deliberate_close = True
                self._connection_auth_timeout = True
                self._client._on_error(
                    constants.topic.CONNECTION, data[0], data[1])

    def _get_auth_data(self, data):
        if data:
            return message_parser.convert_typed(data, self._client)

    def _set_state(self, state):
        self._state = state
        self._client.emit(constants.event.CONNECTION_STATE_CHANGED, state)

    @property
    def state(self):
        """str: State of the connection.

        The possible states are defined in constants.connection_state:
            - closed
            - awaiting_authentication
            - authenticating
            - open
            - error
            - reconnecting
        """
        return self._state

    @property
    def io_loop(self):
        return self._io_loop

    def send_message(self, topic, action, data):
        message = message_builder.get_message(topic, action, data)
        return self.send(message)

    def send(self, raw_message):
        """Main method for sending messages.

        All messages are passed onto and handled by tornado.
        """
        if not self._websocket_handler.stream.closed():
            return self._websocket_handler.write_message(raw_message.encode())
        else:
            future = concurrent.Future()
            self._queued_messages.append((raw_message.encode(), future))
            return future

    @gen.coroutine
    def _send_queued_messages(self):
        if self._state != constants.connection_state.OPEN:
            return

        while self._queued_messages:
            raw_message, future = self._queued_messages.popleft()
            result = yield self._websocket_handler.write_message(raw_message)
            future.set_result(result)

    def _on_data(self, data):
        if data is None:
            self._on_close()
            return
        full_buffer = self._message_buffer + data
        split_buffer = full_buffer.rsplit(constants.message.MESSAGE_SEPERATOR,
                                          1)
        if len(split_buffer) > 1:
            self._message_buffer = split_buffer[1]

        raw_messages = split_buffer[0]

        parsed_messages = message_parser.parse(raw_messages, self._client)

        for msg in parsed_messages:
            if msg is None:
                continue
            if msg['topic'] == constants.topic.CONNECTION:
                self._handle_connection_response(msg)
            elif msg['topic'] == constants.topic.AUTH:
                self._handle_auth_response(msg)
            else:
                self._client._on_message(parsed_messages[0])

    def _try_reconnect(self):
        if self._reconnect_timeout is not None:
            return

        if self._reconnection_attempt < self._max_reconnect_attempts:
            self._set_state(constants.connection_state.RECONNECTING)

            interval = min(
                (self._reconnect_interval_increment *
                 self._reconnection_attempt),
                self._max_reconnect_interval)

            self._reconnection_attempt += 1

            self._reconnect_timeout = self._io_loop.call_later(
                interval, self._try_open)

        else:
            self._clear_reconnect()
            self.close()

    def _try_open(self):
        self._url = self._original_url
        self.connect()
        self._reconnect_timeout = None

    def _clear_reconnect(self):
        self._io_loop.remove_timeout(self._reconnect_timeout)
        self._reconnect_timeout = None
        self._reconnection_attempt = 0

    def _on_close(self):
        self._io_loop.remove_timeout(self._heartbeat_callback)

        if self._redirecting:
            self._redirecting = False
            self.connect()
        elif self._deliberate_close:
            self._set_state(constants.connection_state.CLOSED)
        else:
            self._try_reconnect()



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

import re
from copy import deepcopy
from deepstreampy.utils import Undefined, num_types

SPLIT_REG_EXP = r"[\.\[\]]"


def get(data, path, deep_copy):
    node = data
    tokens = _tokenize(path)
    for token in tokens:
        try:
            node = node[token]
        except (IndexError, KeyError):
            return None

    if deep_copy:
        return deepcopy(node)
    else:
        return node


def set(data, path, value, deep_copy):
    if deep_copy:
        data = deepcopy(data)

    node = data

    tokens = _tokenize(path)
    if not tokens:
        data = value
        return data

    for i, token in enumerate(tokens[:-1]):
        try:
            node = node[token]
        except (KeyError, IndexError):
            if (i + 1 < len(tokens) and
                    isinstance(tokens[i+1], int)):
                if isinstance(node, list):
                    _pad_list(node, token + 1, None)

                node[token] = []
            else:
                if isinstance(node, list):
                    _pad_list(node, token + 1, None)

                node[token] = {}

            node = node[token]
    last_token = tokens[-1]
    if value is Undefined:
        del node[last_token]
        return data

    if isinstance(node, list):
        _pad_list(node, len(node) + 1, None)

    node[last_token] = value

    return data


def _tokenize(path):
    if not path:
        return ()

    if isinstance(path, num_types):
        return [path]
    parts = re.split(SPLIT_REG_EXP, path)
    tokens = []
    for part in parts:
        if len(part) == 0:
            continue
        try:
            tokens.append(int(part))
        except ValueError:
            if part == '*':
                tokens.append(True)
            else:
                tokens.append(part)

    return tokens


def _pad_list(_list, index, value):
    if len(_list) < index - 1:
        _list.extend([value] * (index - len(_list)))



"""Tests for parsing and building messages and converting parsed messages."""
from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.message import message_builder, message_parser
from deepstreampy.constants import topic, actions, types
from deepstreampy.constants import topic as topic_constants
from deepstreampy.constants import event as event_constants
from deepstreampy import client

import json
import unittest

URL = "ws://localhost:7777/deepstream"

class MessageTest(unittest.TestCase):

    def setUp(self):
        # This client will never connect
        self.client = client.Client(URL)

    def test_message_build_and_parse(self):
        """Test building messages and parsing them back correctly."""
        data = [{'username': 'bob'}]
        real_message = message_builder.get_message(topic.AUTH,
                                                   actions.REQUEST,
                                                   data)
        expected_message = (
            topic.AUTH + chr(31) + actions.REQUEST + chr(31) +
            json.dumps(data[0], separators=(',', ':')) + chr(30))
        self.assertEqual(real_message, expected_message)
        parsed_messages = message_parser.parse(real_message, self.client)
        self.assertEqual(len(parsed_messages), 1)
        message = parsed_messages[0]
        self.assertEqual(message['topic'], topic.AUTH)
        self.assertEqual(message['action'], actions.REQUEST)
        self.assertEqual([json.loads(message['data'][0])], data)
        data = ['entry1', 'entry2']
        real_message = message_builder.get_message(topic.EVENT,
                                                   actions.READ,
                                                   data)
        expected_message = (topic.EVENT + chr(31) +
                            actions.READ + chr(31) +
                            data[0] + chr(31) +
                            data[1] + chr(30))
        self.assertEqual(real_message, expected_message)

        parsed_messages = message_parser.parse(real_message, self.client)
        self.assertEqual(len(parsed_messages), 1)

        message = parsed_messages[0]
        self.assertEqual(message['topic'], topic.EVENT)
        self.assertEqual(message['action'], actions.READ)
        self.assertEqual(message['data'], data)

    def test_wrong_action(self):
        """Test parsing message with an action that doesn't exist."""
        self.assertRaises(ValueError,
                          message_parser.parse,
                          topic.AUTH + chr(31) + 'F' + chr(30),
                          self.client)

    def test_parse_short_message(self):
        """Test parsing message with too few parts."""
        self.assertRaises(ValueError,
                          message_parser.parse,
                          topic.AUTH + chr(30),
                          self.client)

    def test_to_typed(self):
        """Test convert to typed."""
        self.assertEqual(message_builder.typed("somestring"),
                         types.STRING + "somestring")
        self.assertEqual(message_builder.typed(1), "N1")
        self.assertEqual(message_builder.typed(2.3), "N2.3")
        self.assertEqual(message_builder.typed(True), "T")
        self.assertEqual(message_builder.typed(False), "F")
        self.assertEqual(message_builder.typed(None), "L")
        object_message = message_builder.typed(
            {"foo": 2, "bar": True, 'a': 'b'})
        self.assertEqual(object_message[0], 'O')
        self.assertTrue('"foo":2' in object_message)
        self.assertTrue('"bar":true' in object_message)
        self.assertTrue('"a":"b"' in object_message)

    def test_from_typed(self):
        """Test convert from typed."""
        self.assertEqual(
            message_parser.convert_typed("Ssomestring", self.client),
            "somestring")

        self.assertEqual(message_parser.convert_typed("N1", self.client), 1)
        self.assertRaises(ValueError,
                          message_parser.convert_typed,
                          "X21323",
                          self.client)

        self.client.once('error', self._handle_unknown_type)
        message_parser.convert_typed("X21323", self.client)

        self.client.once('error', self._handle_faulty_object)
        message_parser.convert_typed("O<21323>", self.client)

    def _handle_unknown_type(self, message, event, topic):
        self.assertEqual(message, "UNKNOWN_TYPE (X21323)")
        self.assertEqual(topic, topic_constants.ERROR)
        self.assertEqual(event, event_constants.MESSAGE_PARSE_ERROR)

    def _handle_faulty_object(self, message, event, topic):
        self.assertTrue("JSON object", message)
        self.assertEqual(topic, topic_constants.ERROR)
        self.assertEqual(event, event_constants.MESSAGE_PARSE_ERROR)

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, division, print_function, with_statement
import sys
import inspect


PING = 'PI'
PONG = 'PO'
ACK = 'A'
REDIRECT = 'RED'
CHALLENGE = 'CH'
CHALLENGE_RESPONSE = 'CHR'
READ = 'R'
CREATE = 'C'
UPDATE = 'U'
PATCH = 'P'
DELETE = 'D'
SUBSCRIBE = 'S'
UNSUBSCRIBE = 'US'
HAS = 'H'
SNAPSHOT = 'SN'
INVOKE = 'I'
SUBSCRIPTION_FOR_PATTERN_FOUND = 'SP'
SUBSCRIPTION_FOR_PATTERN_REMOVED = 'SR'
SUBSCRIPTION_HAS_PROVIDER = 'SH'
LISTEN = 'L'
UNLISTEN = 'UL'
LISTEN_ACCEPT = 'LA'
LISTEN_REJECT = 'LR'
PROVIDER_UPDATE = 'PU'
QUERY = 'Q'
CREATEORREAD = 'CR'
EVENT = 'EVT'
ERROR = 'E'
REQUEST = 'REQ'
RESPONSE = 'RES'
REJECTION = 'REJ'
PRESENCE_JOIN = 'PNJ'
PRESENCE_LEAVE = 'PNL'
QUERY = 'Q'
WRITE_ACKNOWLEDGEMENT = 'WA'


_current_module = sys.modules[__name__]

_reverse_lookup_map = dict()
for name, value in inspect.getmembers(_current_module):
    if name.isupper():
        _reverse_lookup_map[value] = name


def reverse_lookup(action):
    return _reverse_lookup_map.get(action, None)



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from tornado import tcpserver, concurrent


def msg(value):
    return value.replace("|", chr(31)).replace("+", chr(30)).encode()


class FakeServer(tcpserver.TCPServer):

    def __init__(self):
        super(FakeServer, self).__init__()
        self.queued_messages = list()
        self.received_messages = list()
        self.stream = None
        self.connections = []
        self._messsage_future = None

    def write(self, message):
        self.queued_messages.append(message)
        self.stream.write(message.encode())

    def read(self, data):
        message = data.decode()
        self.received_messages.append(message)
        if self._messsage_future:
            self._messsage_future.set_result(message)

    def handle_stream(self, stream, address):
        self.connections.append(stream)
        self.stream = stream
        self.stream.read_until_close(None, self.read)

    def reset_message_count(self):
        self.received_messages = list()

    def await_message(self):
        self._messsage_future = concurrent.Future()
        return self._messsage_future



# coding: utf8
################################################################################
#
#   Autore          :   Allan Nava
#   Modificato      :   Allan Nava
#   Data            :   18/03/2020
#   Aggiornamento   :   18/03/2020
#
#################################################################################
#
import os
from setuptools import setup, find_packages
#
def requirements(filename='requirements.txt'):
    with open(filename) as f:
        requirements = f.read().splitlines()
        return requirements
#
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()
#
setup(name='deepstreampy',
      version='0.1.3',
      author='Yavor Paunov - Allan Nava',
      author_email='contact@yavorpaunov.com',
      description='A deepstream.io client.',
      license='MIT',
      url='https://github.com/Allan-Nava/deepstreampy',
      packages=find_packages(),
      download_url = 'https://github.com/Allan-Nava/deepstreampy/archive/v0.1.3.tar.gz', 
      classifiers=[
        'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',      # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
      ],
      install_requires=requirements(),
      long_description=read('README.md'),
      test_suite='tests')
#


from __future__ import absolute_import, division, print_function, with_statement
from deepstreampy.constants import types
from deepstreampy.constants import message as message_constants
from deepstreampy.utils import Undefined
import sys
import json


def get_message(topic, action, data=None):
    send_data = [topic, action]

    if data:
        for param in data:
            if isinstance(param, dict):
                value = json.dumps(param,
                                   separators=(',', ':'),
                                   sort_keys=True)
                send_data.append(value)
            elif isinstance(param, list):
                value = ("[" +
                         ",".join('"{0}"'.format(item) for item in param) +
                         "]")
                send_data.append(value)
            else:
                send_data.append(str(param))

    full_message = (message_constants.MESSAGE_PART_SEPERATOR.join(send_data) +
                    message_constants.MESSAGE_SEPERATOR)
    return full_message


def typed(value):
    if value is None:
        return types.NULL

    value_type = type(value)

    if sys.version_info < (3,):
        num_types = (int, long, float, complex)
        str_types = (str, unicode)
    else:
        num_types = (int, float, complex)
        str_types = (str,)
    obj_types = (dict, list,)

    if value_type in str_types:
        return types.STRING + value

    if value_type in obj_types:
        return types.OBJECT + json.dumps(value,
                                         separators=(',', ':'),
                                         sort_keys=False)

    if value_type is bool:
        if value:
            return types.TRUE
        else:
            return types.FALSE

    if value_type in num_types:
        return types.NUMBER + str(value)

    if value is Undefined:
        return types.UNDEFINED

    raise ValueError("Can't serialize type {0}".format(value_type))
